from __future__ import annotations

from chik.util.ints import uint32, uint64

# 1 Chik coin = 1,000,000,000,000 = 1 trillion mojo.
_mojo_per_chik = 1000000000000
_blocks_per_year = 1681920  # 32 * 6 * 24 * 365


def calculate_pool_reward(height: uint32) -> uint64:
    """
    Returns the pool reward at a certain block height. The pool earns 7/8 of the reward in each block. If the farmer
    is solo farming, they act as the pool, and therefore earn the entire block reward.
    These halving events will not be hit at the exact times
    (3 years, etc), due to fluctuations in difficulty. They will likely come early, if the network space and VDF
    rates increase continuously.
    """

    if height == 0:
        return uint64(int((7 / 8) * 21000000 * _mojo_per_chik))
    elif height < 3 * _blocks_per_year:
        return uint64(int((7 / 8) * 2 * _mojo_per_chik))
    elif height < 6 * _blocks_per_year:
        return uint64(int((7 / 8) * 1 * _mojo_per_chik))
    elif height < 9 * _blocks_per_year:
        return uint64(int((7 / 8) * 0.5 * _mojo_per_chik))
    elif height < 12 * _blocks_per_year:
        return uint64(int((7 / 8) * 0.25 * _mojo_per_chik))
    else:
        return uint64(int((7 / 8) * 0.125 * _mojo_per_chik))


def calculate_base_farmer_reward(height: uint32) -> uint64:
    """
    Returns the base farmer reward at a certain block height.
    The base fee reward is 1/8 of total block reward

    Returns the coinbase reward at a certain block height. These halving events will not be hit at the exact times
    (3 years, etc), due to fluctuations in difficulty. They will likely come early, if the network space and VDF
    rates increase continuously.
    """
    if height == 0:
        return uint64(int((1 / 8) * 21000000 * _mojo_per_chik))
    elif height < 3 * _blocks_per_year:
        return uint64(int((1 / 8) * 2 * _mojo_per_chik))
    elif height < 6 * _blocks_per_year:
        return uint64(int((1 / 8) * 1 * _mojo_per_chik))
    elif height < 9 * _blocks_per_year:
        return uint64(int((1 / 8) * 0.5 * _mojo_per_chik))
    elif height < 12 * _blocks_per_year:
        return uint64(int((1 / 8) * 0.25 * _mojo_per_chik))
    else:
        return uint64(int((1 / 8) * 0.125 * _mojo_per_chik))



from __future__ import annotations

from typing import Optional

import chik_rs
from typing_extensions import Protocol

from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32, uint64

BlockRecord = chik_rs.BlockRecord


class BlockRecordProtocol(Protocol):
    @property
    def header_hash(self) -> bytes32: ...

    @property
    def height(self) -> uint32: ...

    @property
    def timestamp(self) -> Optional[uint64]: ...

    @property
    def prev_transaction_block_height(self) -> uint32: ...

    @property
    def prev_transaction_block_hash(self) -> Optional[bytes32]: ...

    @property
    def is_transaction_block(self) -> bool: ...



from __future__ import annotations

import re

import pytest
from chik_rs import AugSchemeMPL, G1Element, G2Element, PrivateKey

from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, make_spend
from chik.types.condition_opcodes import ConditionOpcode
from chik.util.db_wrapper import DBWrapper2, manage_connection
from chik.util.ints import uint32, uint64
from chik.wallet.derivation_record import DerivationRecord
from chik.wallet.derive_keys import master_sk_to_wallet_sk, master_sk_to_wallet_sk_unhardened
from chik.wallet.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
    DEFAULT_HIDDEN_PUZZLE_HASH,
    calculate_synthetic_secret_key,
    puzzle_hash_for_synthetic_public_key,
)
from chik.wallet.util.wallet_types import WalletType
from chik.wallet.wallet import Wallet
from chik.wallet.wallet_puzzle_store import WalletPuzzleStore
from chik.wallet.wallet_state_manager import WalletStateManager
from chik.wallet.wallet_user_store import WalletUserStore

top_sk: PrivateKey = PrivateKey.from_bytes(bytes([1] * 32))
sk1_h: PrivateKey = master_sk_to_wallet_sk(top_sk, uint32(1))
sk2_h: PrivateKey = master_sk_to_wallet_sk(top_sk, uint32(2))
sk2_h_synth: PrivateKey = calculate_synthetic_secret_key(sk2_h, DEFAULT_HIDDEN_PUZZLE_HASH)
sk1_u: PrivateKey = master_sk_to_wallet_sk_unhardened(top_sk, uint32(1))
sk2_u: PrivateKey = master_sk_to_wallet_sk(top_sk, uint32(2))
sk2_u_synth: PrivateKey = calculate_synthetic_secret_key(sk2_u, DEFAULT_HIDDEN_PUZZLE_HASH)
pk1_h: G1Element = sk1_h.get_g1()
pk2_h: G1Element = sk2_h.get_g1()
pk2_h_synth: G1Element = sk2_h_synth.get_g1()
pk1_u: G1Element = sk1_u.get_g1()
pk2_u: G1Element = sk2_u.get_g1()
pk2_u_synth: G1Element = sk2_u_synth.get_g1()
msg1: bytes = b"msg1"
msg2: bytes = b"msg2"

additional_data: bytes32 = bytes32(DEFAULT_CONSTANTS.AGG_SIG_ME_ADDITIONAL_DATA)

coin: Coin = Coin(bytes32.zeros, bytes32.zeros, uint64(0))
puzzle = SerializedProgram.from_bytes(b"\x01")
solution_h = SerializedProgram.from_program(
    Program.to([[ConditionOpcode.AGG_SIG_UNSAFE, pk1_h, msg1], [ConditionOpcode.AGG_SIG_ME, pk2_h_synth, msg2]])
)
solution_u = SerializedProgram.from_program(
    Program.to([[ConditionOpcode.AGG_SIG_UNSAFE, pk1_u, msg1], [ConditionOpcode.AGG_SIG_ME, pk2_u_synth, msg2]])
)
spend_h: CoinSpend = make_spend(
    coin,
    puzzle,
    solution_h,
)
spend_u: CoinSpend = make_spend(
    coin,
    puzzle,
    solution_u,
)


@pytest.mark.anyio
async def test_wsm_sign_transaction() -> None:
    async with manage_connection("file:temp.db?mode=memory&cache=shared", uri=True, name="writer") as writer_conn:
        async with manage_connection("file:temp.db?mode=memory&cache=shared", uri=True, name="reader") as reader_conn:
            wsm = WalletStateManager()
            db = DBWrapper2(writer_conn)
            await db.add_connection(reader_conn)
            wsm.puzzle_store = await WalletPuzzleStore.create(db)
            wsm.constants = DEFAULT_CONSTANTS
            wsm.private_key = top_sk
            wsm.root_pubkey = top_sk.get_g1()
            wsm.user_store = await WalletUserStore.create(db)
            wallet_info = await wsm.user_store.get_wallet_by_id(1)
            assert wallet_info is not None
            wsm.main_wallet = await Wallet.create(wsm, wallet_info)

            with pytest.raises(
                ValueError, match=re.escape(f"Pubkey {pk1_h.get_fingerprint()} not found (or path/sum hinted to)")
            ):
                await wsm.sign_bundle([spend_h])

            await wsm.puzzle_store.add_derivation_paths(
                [
                    DerivationRecord(
                        uint32(1),
                        bytes32.zeros,
                        pk1_h,
                        WalletType.STANDARD_WALLET,
                        uint32(1),
                        True,
                    )
                ]
            )

            await wsm.puzzle_store.add_derivation_paths(
                [
                    DerivationRecord(
                        uint32(2),
                        puzzle_hash_for_synthetic_public_key(pk2_h_synth),
                        pk2_h,
                        WalletType.STANDARD_WALLET,
                        uint32(1),
                        True,
                    )
                ]
            )

            signature: G2Element = ((await wsm.sign_bundle([spend_h]))[0]).aggregated_signature
            assert signature == AugSchemeMPL.aggregate(
                [
                    AugSchemeMPL.sign(sk1_h, msg1),
                    AugSchemeMPL.sign(sk2_h_synth, msg2 + coin.name() + additional_data),
                ]
            )

            with pytest.raises(
                ValueError, match=re.escape(f"Pubkey {pk1_u.get_fingerprint()} not found (or path/sum hinted to)")
            ):
                await wsm.sign_bundle([spend_u])

            await wsm.puzzle_store.add_derivation_paths(
                [
                    DerivationRecord(
                        uint32(1),
                        bytes32.zeros,
                        pk1_u,
                        WalletType.STANDARD_WALLET,
                        uint32(1),
                        False,
                    )
                ]
            )

            await wsm.puzzle_store.add_derivation_paths(
                [
                    DerivationRecord(
                        uint32(2),
                        puzzle_hash_for_synthetic_public_key(pk2_u_synth),
                        pk2_u,
                        WalletType.STANDARD_WALLET,
                        uint32(1),
                        False,
                    )
                ]
            )
            signature2: G2Element = ((await wsm.sign_bundle([spend_u]))[0]).aggregated_signature
            assert signature2 == AugSchemeMPL.aggregate(
                [
                    AugSchemeMPL.sign(sk1_u, msg1),
                    AugSchemeMPL.sign(sk2_u_synth, msg2 + coin.name() + additional_data),
                ]
            )



from __future__ import annotations

from chik.full_node.full_node_api import FullNodeAPI
from chik.util.ints import uint32


def node_height_at_least(node: FullNodeAPI, h: uint32) -> bool:
    if node.full_node.blockchain.get_peak() is not None:
        peak = node.full_node.blockchain.get_peak()
        if peak is not None:
            return peak.height >= h
    return False


def node_height_exactly(node: FullNodeAPI, h: uint32) -> bool:
    if node.full_node.blockchain.get_peak() is not None:
        peak = node.full_node.blockchain.get_peak()
        if peak is not None:
            return peak.height == h
    return False


def node_height_between(node: FullNodeAPI, h1: uint32, h2: uint32) -> bool:
    if node.full_node.blockchain.get_peak() is not None:
        peak = node.full_node.blockchain.get_peak()
        if peak is not None:
            return h1 <= peak.height <= h2
    return False



from __future__ import annotations

import asyncio
import copy
import logging
import time
import traceback
from collections.abc import Awaitable, Collection
from concurrent.futures import Executor
from dataclasses import dataclass
from typing import Optional

from chik_rs import SpendBundleConditions, get_flags_for_height_and_constants, run_block_generator, run_block_generator2

from chik.consensus.block_header_validation import validate_finished_header_block
from chik.consensus.block_record import BlockRecord
from chik.consensus.blockchain_interface import BlockRecordsProtocol
from chik.consensus.constants import ConsensusConstants
from chik.consensus.full_block_to_block_record import block_to_block_record
from chik.consensus.get_block_challenge import get_block_challenge
from chik.consensus.get_block_generator import get_block_generator
from chik.consensus.pot_iterations import calculate_iterations_quality, is_overflow_block
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.proof_of_space import verify_and_get_quality_string
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.blockchain_format.sub_epoch_summary import SubEpochSummary
from chik.types.full_block import FullBlock
from chik.types.generator_types import BlockGenerator
from chik.types.validation_state import ValidationState
from chik.util.augmented_chain import AugmentedBlockchain
from chik.util.errors import Err
from chik.util.generator_tools import get_block_header, tx_removals_and_additions
from chik.util.ints import uint16, uint32, uint64
from chik.util.streamable import Streamable, streamable

log = logging.getLogger(__name__)


@streamable
@dataclass(frozen=True)
class PreValidationResult(Streamable):
    error: Optional[uint16]
    required_iters: Optional[uint64]  # Iff error is None
    conds: Optional[SpendBundleConditions]  # Iff error is None and block is a transaction block
    timing: uint32  # the time (in milliseconds) it took to pre-validate the block

    @property
    def validated_signature(self) -> bool:
        if self.conds is None:
            return False
        return self.conds.validated_signature


# this layer of abstraction is here to let wallet tests monkeypatch it
def _run_block(
    block: FullBlock, prev_generators: list[bytes], constants: ConsensusConstants
) -> tuple[Optional[int], Optional[SpendBundleConditions]]:
    assert block.transactions_generator is not None
    assert block.transactions_info is not None
    flags = get_flags_for_height_and_constants(block.height, constants)
    if block.height >= constants.HARD_FORK_HEIGHT:
        run_block = run_block_generator2
    else:
        run_block = run_block_generator
    return run_block(
        bytes(block.transactions_generator),
        prev_generators,
        block.transactions_info.cost,
        flags,
        block.transactions_info.aggregated_signature,
        None,
        constants,
    )


def _pre_validate_block(
    constants: ConsensusConstants,
    blockchain: BlockRecordsProtocol,
    block: FullBlock,
    prev_generators: Optional[list[bytes]],
    conds: Optional[SpendBundleConditions],
    expected_vs: ValidationState,
) -> PreValidationResult:
    """
    Args:
        constants:
        blockchain:
        block:
        prev_generators:
        conds:
        expected_vs: The validation state that we calculate for the next block
            if it's validated.
    """

    try:
        validation_start = time.monotonic()
        removals_and_additions: Optional[tuple[Collection[bytes32], Collection[Coin]]] = None
        if conds is not None:
            assert conds.validated_signature is True
            assert block.transactions_generator is not None
            removals_and_additions = tx_removals_and_additions(conds)
        elif block.transactions_generator is not None:
            assert prev_generators is not None
            assert block.transactions_info is not None

            if block.transactions_info.cost > constants.MAX_BLOCK_COST_KLVM:
                validation_time = time.monotonic() - validation_start
                return PreValidationResult(
                    uint16(Err.BLOCK_COST_EXCEEDS_MAX.value), None, None, uint32(validation_time * 1000)
                )

            err, conds = _run_block(block, prev_generators, constants)

            assert (err is None) != (conds is None)
            if err is not None:
                validation_time = time.monotonic() - validation_start
                return PreValidationResult(uint16(err), None, None, uint32(validation_time * 1000))
            assert conds is not None
            assert conds.validated_signature is True
            removals_and_additions = tx_removals_and_additions(conds)
        elif block.is_transaction_block():
            # This is a transaction block with just reward coins.
            removals_and_additions = ([], [])

        assert conds is None or conds.validated_signature is True
        header_block = get_block_header(block, removals_and_additions)
        required_iters, error = validate_finished_header_block(
            constants,
            blockchain,
            header_block,
            True,  # check_filter
            expected_vs,
        )
        error_int: Optional[uint16] = None
        if error is not None:
            error_int = uint16(error.code.value)

        validation_time = time.monotonic() - validation_start
        return PreValidationResult(
            error_int,
            required_iters,
            conds,
            uint32(validation_time * 1000),
        )
    except Exception:
        error_stack = traceback.format_exc()
        log.error(f"Exception: {error_stack}")
        validation_time = time.monotonic() - validation_start
        return PreValidationResult(uint16(Err.UNKNOWN.value), None, None, uint32(validation_time * 1000))


async def pre_validate_block(
    constants: ConsensusConstants,
    blockchain: AugmentedBlockchain,
    block: FullBlock,
    pool: Executor,
    conds: Optional[SpendBundleConditions],
    vs: ValidationState,
    *,
    wp_summaries: Optional[list[SubEpochSummary]] = None,
) -> Awaitable[PreValidationResult]:
    """
    This method must be called under the blockchain lock
    The block passed to this function is submitted to be validated in the
    executor passed in as "pool". The future for the job is then returned.
    When awaited, the return value is the PreValidationResult for the block.
    The PreValidationResult indicates whether the block was valid or not.

    Args:
        constants:
        blockchain: The blockchain object to validate these blocks with respect to.
            It's an AugmentedBlockchain to allow for previous batches of blocks to
            be included, even if they haven't been added to the underlying blockchain
            database yet. The blocks passed in will be added/augmented onto this blockchain.
        pool: The executor to submit the validation jobs to
        block: The full block to validate (must be connected to current chain)
        conds: The SpendBundleConditions for transaction blocks, if we have one.
            This will be computed if None is passed.
        vs: The ValidationState refers to the state for the block.
            This is an in-out parameter that will be updated to the validation state
            for the next block. It includes subslot iterators, difficulty and
            the previous sub epoch summary (ses) block.
        wp_summaries:
        validate_signatures:
    """
    prev_b: Optional[BlockRecord] = None

    async def return_error(error_code: Err) -> PreValidationResult:
        return PreValidationResult(uint16(error_code.value), None, None, uint32(0))

    if block.height > 0:
        curr = blockchain.try_block_record(block.prev_header_hash)
        if curr is None:
            return return_error(Err.INVALID_PREV_BLOCK_HASH)
        prev_b = curr

    assert isinstance(block, FullBlock)
    if len(block.finished_sub_slots) > 0:
        if block.finished_sub_slots[0].challenge_chain.new_difficulty is not None:
            vs.difficulty = block.finished_sub_slots[0].challenge_chain.new_difficulty
        if block.finished_sub_slots[0].challenge_chain.new_sub_slot_iters is not None:
            vs.ssi = block.finished_sub_slots[0].challenge_chain.new_sub_slot_iters
    overflow = is_overflow_block(constants, block.reward_chain_block.signage_point_index)
    challenge = get_block_challenge(constants, block, blockchain, prev_b is None, overflow, False)
    if block.reward_chain_block.challenge_chain_sp_vdf is None:
        cc_sp_hash: bytes32 = challenge
    else:
        cc_sp_hash = block.reward_chain_block.challenge_chain_sp_vdf.output.get_hash()
    q_str: Optional[bytes32] = verify_and_get_quality_string(
        block.reward_chain_block.proof_of_space, constants, challenge, cc_sp_hash, height=block.height
    )
    if q_str is None:
        return return_error(Err.INVALID_POSPACE)

    required_iters: uint64 = calculate_iterations_quality(
        constants.DIFFICULTY_CONSTANT_FACTOR,
        q_str,
        block.reward_chain_block.proof_of_space.size,
        vs.difficulty,
        cc_sp_hash,
    )

    try:
        block_rec = block_to_block_record(
            constants,
            blockchain,
            required_iters,
            block,
            sub_slot_iters=vs.ssi,
            prev_ses_block=vs.prev_ses_block,
        )
    except ValueError:
        log.exception("block_to_block_record()")
        return return_error(Err.INVALID_SUB_EPOCH_SUMMARY)

    if block_rec.sub_epoch_summary_included is not None and wp_summaries is not None:
        next_ses = wp_summaries[int(block.height / constants.SUB_EPOCH_BLOCKS) - 1]
        if not block_rec.sub_epoch_summary_included.get_hash() == next_ses.get_hash():
            log.error("sub_epoch_summary does not match wp sub_epoch_summary list")
            return return_error(Err.INVALID_SUB_EPOCH_SUMMARY)

    blockchain.add_extra_block(block, block_rec)  # Temporarily add block to chain
    prev_b = block_rec

    previous_generators: Optional[list[bytes]] = None

    try:
        block_generator: Optional[BlockGenerator] = await get_block_generator(blockchain.lookup_block_generators, block)
        if block_generator is not None:
            previous_generators = block_generator.generator_refs
    except ValueError:
        return return_error(Err.FAILED_GETTING_GENERATOR_MULTIPROCESSING)

    future = asyncio.get_running_loop().run_in_executor(
        pool,
        _pre_validate_block,
        constants,
        blockchain,
        block,
        previous_generators,
        conds,
        copy.copy(vs),
    )

    if block_rec.sub_epoch_summary_included is not None:
        vs.prev_ses_block = block_rec

    return future



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Any, Optional

import pytest
from chik_rs import AugSchemeMPL, G1Element, G2Element

from chik._tests.environments.wallet import WalletStateTransition, WalletTestFramework
from chik._tests.util.time_out_assert import time_out_assert
from chik.rpc.wallet_request_types import GetTransactionMemo
from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.simulator.simulator_protocol import ReorgProtocol
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, compute_additions
from chik.types.peer_info import PeerInfo
from chik.types.signing_mode import CHIP_0002_SIGN_MESSAGE_PREFIX
from chik.types.spend_bundle import estimate_fees
from chik.util.bech32m import encode_puzzle_hash
from chik.util.errors import Err
from chik.util.ints import uint16, uint32, uint64
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.derive_keys import master_sk_to_wallet_sk
from chik.wallet.payment import Payment
from chik.wallet.puzzles.clawback.metadata import AutoClaimSettings
from chik.wallet.transaction_record import TransactionRecord
from chik.wallet.util.query_filter import TransactionTypeFilter
from chik.wallet.util.transaction_type import TransactionType
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG
from chik.wallet.util.wallet_types import CoinType
from chik.wallet.wallet_node import WalletNode, get_wallet_db_path


class TestWalletSimulator:
    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [10], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_coinbase(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wsm = env.wallet_state_manager

        all_txs = await wsm.get_all_transactions(1)
        assert len(all_txs) == 20

        pool_rewards = 0
        farm_rewards = 0

        for tx in all_txs:
            if TransactionType(tx.type) == TransactionType.COINBASE_REWARD:
                pool_rewards += 1
            elif TransactionType(tx.type) == TransactionType.FEE_REWARD:
                farm_rewards += 1

        assert pool_rewards == 10
        assert farm_rewards == 10

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [1]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_make_transaction(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        tx_amount = 10

        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                bytes32.zeros,
                action_scope,
                uint64(0),
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,
                            ">=#max_send_amount": 1,
                            "<=#pending_change": 1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                )
            ]
        )

        # Test match_hinted_coin
        async with wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=False
        ) as action_scope:
            selected_coin = next(iter(await wallet.select_coins(uint64(0), action_scope)))
        assert await wallet.match_hinted_coin(selected_coin, selected_coin.puzzle_hash)

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [1], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_reuse_address(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        tx_amount = 10

        async with wallet.wallet_state_manager.new_action_scope(
            DEFAULT_TX_CONFIG.override(reuse_puzhash=True), push=True
        ) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                bytes32.zeros,
                action_scope,
                uint64(0),
            )
        [tx] = action_scope.side_effects.transactions
        assert tx.spend_bundle is not None
        assert len(tx.spend_bundle.coin_spends) == 1
        new_puzhash = [c.puzzle_hash.hex() for c in tx.additions]
        assert tx.spend_bundle.coin_spends[0].coin.puzzle_hash.hex() in new_puzhash
        [tx] = await wallet.wallet_state_manager.add_pending_transactions([tx])

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,
                            ">=#max_send_amount": 1,
                            "<=#pending_change": 1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                )
            ]
        )

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [2, 1], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.parametrize("number_of_coins", [1, 3])
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_clawback_claim_auto(
        self, wallet_environments: WalletTestFramework, number_of_coins: int
    ) -> None:
        env = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet = env.xck_wallet
        wallet_1 = env_1.xck_wallet
        wsm = env.wallet_state_manager
        wsm_1 = env_1.wallet_state_manager

        tx_amount = 500
        normal_puzhash = await wallet_1.get_new_puzzlehash()

        # Transfer to normal wallet
        for _ in range(0, number_of_coins):
            async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
                await wallet.generate_signed_transaction(
                    uint64(tx_amount),
                    normal_puzhash,
                    action_scope,
                    uint64(0),
                    puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 10}],
                )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount * number_of_coins,
                            "<=#spendable_balance": -1 * tx_amount * number_of_coins,
                            "<=#max_send_amount": -1 * tx_amount * number_of_coins,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": number_of_coins,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount * number_of_coins,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -number_of_coins,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, number_of_coins, tx_amount * 2, CoinType.CLAWBACK)
        await time_out_assert(
            20, wsm_1.coin_store.count_small_unspent, number_of_coins, tx_amount * 2, CoinType.CLAWBACK
        )

        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                normal_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 10}],
            )
        [tx_bad] = action_scope.side_effects.transactions

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Change one coin to test missing metadata case
        clawback_coin_id = tx_bad.additions[0].name()
        coin_record = await wsm_1.coin_store.get_coin_record(clawback_coin_id)
        assert coin_record is not None
        await wsm_1.coin_store.add_coin_record(dataclasses.replace(coin_record, metadata=None))
        # Claim merkle coin
        env_1.node.set_auto_claim(AutoClaimSettings(enabled=True, batch_size=uint16(2)))
        # Trigger auto claim
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    # After auto claim is set, the next block will trigger submission of clawback claims
                    post_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": tx_amount * number_of_coins,
                            "pending_change": tx_amount
                            * number_of_coins,  # This is a little weird but I think intentional and correct
                            "pending_coin_removal_count": number_of_coins,
                        }
                    },
                ),
            ]
        )
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount * number_of_coins,
                            "spendable_balance": tx_amount * number_of_coins,
                            "max_send_amount": tx_amount * number_of_coins,
                            "unspent_coin_count": number_of_coins,
                            "pending_change": -tx_amount * number_of_coins,
                            "pending_coin_removal_count": -1 * number_of_coins,
                        }
                    },
                ),
            ]
        )
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, tx_amount * 2, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_1.coin_store.count_small_unspent, 1, tx_amount * 2, CoinType.CLAWBACK)

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_clawback_clawback(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        env_2 = wallet_environments.environments[1]
        wsm = env.wallet_state_manager
        wsm_2 = env_2.wallet_state_manager
        wallet = env.xck_wallet
        wallet_1 = env_2.xck_wallet
        api_0 = env.rpc_api
        api_1 = env_2.rpc_api

        tx_amount = 500
        normal_puzhash = await wallet_1.get_new_puzzlehash()
        # Transfer to normal wallet
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                normal_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 500}],
                memos=[b"Test"],
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        txs = await api_0.get_transactions(
            dict(type_filter={"values": [TransactionType.INCOMING_CLAWBACK_SEND], "mode": 1}, wallet_id=1)
        )
        # clawback merkle coin
        [tx] = action_scope.side_effects.transactions
        merkle_coin = tx.additions[0] if tx.additions[0].amount == tx_amount else tx.additions[1]
        interested_coins = await wsm_2.interested_store.get_interested_coin_ids()
        assert merkle_coin.name() in set(interested_coins)
        assert len(txs["transactions"]) == 1
        assert not txs["transactions"][0]["confirmed"]
        assert txs["transactions"][0]["metadata"]["recipient_puzzle_hash"][2:] == normal_puzhash.hex()
        assert txs["transactions"][0]["metadata"]["coin_id"] == merkle_coin.name().hex()
        with pytest.raises(ValueError):
            await api_0.spend_clawback_coins({})

        test_fee = 10
        resp = await api_0.spend_clawback_coins(
            {"coin_ids": [normal_puzhash.hex(), merkle_coin.name().hex()], "fee": test_fee}
        )
        assert resp["success"]
        assert len(resp["transaction_ids"]) == 1

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": tx_amount - test_fee,
                            "<=#spendable_balance": -1,
                            "<=#max_send_amount": -1,
                            ">=#pending_change": 1,
                            "pending_coin_removal_count": 2,  # 1 for fee, one for clawback
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount - test_fee,
                            ">=#spendable_balance": 1,
                            ">=#max_send_amount": 1,
                            "<=#pending_change": -1,
                            "unspent_coin_count": 1,
                            "pending_coin_removal_count": -2,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        txs = await api_0.get_transactions(
            dict(
                type_filter={
                    "values": [TransactionType.INCOMING_CLAWBACK_SEND.value, TransactionType.OUTGOING_CLAWBACK.value],
                    "mode": 1,
                },
                wallet_id=1,
            )
        )
        assert len(txs["transactions"]) == 2
        assert txs["transactions"][0]["confirmed"]
        assert txs["transactions"][1]["confirmed"]

        txs = await api_1.get_transactions(
            dict(
                type_filter={
                    "values": [
                        TransactionType.INCOMING_CLAWBACK_RECEIVE.value,
                        TransactionType.OUTGOING_CLAWBACK.value,
                    ],
                    "mode": 1,
                },
                wallet_id=1,
            )
        )
        assert len(txs["transactions"]) == 1
        assert txs["transactions"][0]["confirmed"]
        interested_coins = await wsm_2.interested_store.get_interested_coin_ids()
        assert merkle_coin.name() not in set(interested_coins)

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [1], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_clawback_sent_self(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wsm = env.wallet_state_manager
        wallet = env.xck_wallet
        api_0 = env.rpc_api

        tx_amount = 500
        normal_puzhash = await wallet.get_new_puzzlehash()
        # Transfer to normal wallet
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                normal_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 5}],
                memos=[b"Test"],
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        # Claim merkle coin
        [tx] = action_scope.side_effects.transactions
        merkle_coin = tx.additions[0] if tx.additions[0].amount == tx_amount else tx.additions[1]
        test_fee = 10
        resp = await api_0.spend_clawback_coins(
            {"coin_ids": [merkle_coin.name().hex(), normal_puzhash.hex()], "fee": test_fee}
        )
        assert resp["success"]
        assert len(resp["transaction_ids"]) == 1
        # Wait mempool update
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": tx_amount - test_fee,
                            "<=#spendable_balance": -1,
                            "<=#max_send_amount": -1,
                            ">=#pending_change": 1,
                            "pending_coin_removal_count": 2,  # 1 for fee, one for clawback
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount - test_fee,
                            ">=#spendable_balance": 1,
                            ">=#max_send_amount": 1,
                            "<=#pending_change": -1,
                            "unspent_coin_count": 1,
                            "pending_coin_removal_count": -2,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)

        txs = await api_0.get_transactions(
            dict(
                type_filter={
                    "values": [TransactionType.INCOMING_CLAWBACK_SEND.value, TransactionType.OUTGOING_CLAWBACK.value],
                    "mode": 1,
                },
                wallet_id=1,
            )
        )
        assert len(txs["transactions"]) == 2
        assert txs["transactions"][0]["confirmed"]
        assert txs["transactions"][1]["confirmed"]
        assert txs["transactions"][0]["memos"] != txs["transactions"][1]["memos"]
        assert next(iter(txs["transactions"][0]["memos"].values())) == b"Test".hex()

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_clawback_claim_manual(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        env_2 = wallet_environments.environments[1]
        wsm = env.wallet_state_manager
        wsm_2 = env_2.wallet_state_manager
        wallet = env.xck_wallet
        wallet_1 = env_2.xck_wallet
        api_0 = env.rpc_api
        api_1 = env_2.rpc_api

        tx_amount = 500
        normal_puzhash = await wallet_1.get_new_puzzlehash()
        # Transfer to normal wallet
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                normal_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 5}],
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)

        # Farm a block to pass timelock
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Claim merkle coin
        [tx] = action_scope.side_effects.transactions
        merkle_coin = tx.additions[0] if tx.additions[0].amount == tx_amount else tx.additions[1]
        test_fee = 10
        resp = await api_1.spend_clawback_coins(
            {"coin_ids": [merkle_coin.name().hex(), normal_puzhash.hex()], "fee": test_fee}
        )
        assert resp["success"]
        assert len(resp["transaction_ids"]) == 1

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": tx_amount - test_fee,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 2,  # 1 for fee, 1 for clawback
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount - test_fee,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "unspent_coin_count": 1,
                            "pending_coin_removal_count": -2,
                        }
                    },
                ),
            ]
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)

        txs = await api_0.get_transactions(
            dict(
                type_filter={
                    "values": [
                        TransactionType.INCOMING_CLAWBACK_SEND.value,
                    ],
                    "mode": 1,
                },
                wallet_id=1,
            )
        )
        assert len(txs["transactions"]) == 1
        assert txs["transactions"][0]["confirmed"]

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1], "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_clawback_reorg(self, wallet_environments: WalletTestFramework) -> None:
        full_node_api = wallet_environments.full_node
        env = wallet_environments.environments[0]
        env_2 = wallet_environments.environments[1]
        wsm = env.wallet_state_manager
        wsm_2 = env_2.wallet_state_manager
        wallet = env.xck_wallet
        wallet_1 = env_2.xck_wallet

        tx_amount = 500
        normal_puzhash = await wallet_1.get_new_puzzlehash()
        # Transfer to normal wallet
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                normal_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 5}],
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        # Reorg before claim
        # Test Reorg mint
        height = full_node_api.full_node.blockchain.get_peak_height()
        assert height is not None
        await full_node_api.reorg_from_index_to_new_index(
            ReorgProtocol(uint32(height - 2), uint32(height + 1), bytes32.zeros, None)
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,  # confirmed balance comes back
                            # clawback transaction is now outstanding
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)

        # Claim merkle coin
        env_2.node.set_auto_claim(AutoClaimSettings(enabled=True))
        # clawback merkle coin
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    # After auto claim is set, the next block will trigger submission of clawback claims
                    post_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": tx_amount,
                            "pending_change": tx_amount,  # This is a little weird but I think intentional and correct
                            "pending_coin_removal_count": 1,
                        }
                    },
                ),
            ]
        )
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                            "pending_change": -1 * tx_amount,
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
            ]
        )
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        # Reorg after claim
        height = full_node_api.full_node.blockchain.get_peak_height()
        assert height is not None
        await full_node_api.reorg_from_index_to_new_index(
            ReorgProtocol(uint32(height - 1), uint32(height + 1), bytes32.zeros, None)
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            "spendable_balance": -1 * tx_amount,
                            "max_send_amount": -1 * tx_amount,
                            "unspent_coin_count": -1,
                            "pending_change": tx_amount,
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                            "pending_change": -1 * tx_amount,
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
            ]
        )

        await time_out_assert(20, wsm.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [1], "trusted": True, "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_get_clawback_coins(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wsm = env.wallet_state_manager
        wallet = env.xck_wallet

        tx_amount = 500
        # Transfer to normal wallet
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                bytes32.zeros,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 500}],
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        # clawback merkle coin
        [tx] = action_scope.side_effects.transactions
        merkle_coin = tx.additions[0] if tx.additions[0].amount == tx_amount else tx.additions[1]
        resp = await env.rpc_api.get_coin_records({"wallet_id": 1, "coin_type": 1})
        assert len(resp["coin_records"]) == 1
        assert resp["coin_records"][0]["id"][2:] == merkle_coin.name().hex()

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1], "reuse_puzhash": False}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_clawback_resync(self, self_hostname: str, wallet_environments: WalletTestFramework) -> None:
        full_node_api = wallet_environments.full_node
        env_1 = wallet_environments.environments[0]
        env_2 = wallet_environments.environments[1]
        wsm_1 = env_1.wallet_state_manager
        wsm_2 = env_2.wallet_state_manager
        wallet_1 = env_1.xck_wallet
        wallet_2 = env_2.xck_wallet
        api_1 = env_1.rpc_api

        wallet_1_puzhash = await wallet_1.get_new_puzzlehash()
        wallet_2_puzhash = await wallet_2.get_new_puzzlehash()

        tx_amount = 500
        # Transfer to normal wallet
        async with wallet_1.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_1.generate_signed_transaction(
                uint64(tx_amount),
                wallet_2_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 5}],
            )

        [tx1] = action_scope.side_effects.transactions
        clawback_coin_id_1 = tx1.additions[0].name()
        assert tx1.spend_bundle is not None

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm_1.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)

        tx_amount2 = 700
        async with wallet_1.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_1.generate_signed_transaction(
                uint64(tx_amount2),
                wallet_1_puzhash,
                action_scope,
                uint64(0),
                puzzle_decorator_override=[{"decorator": "CLAWBACK", "clawback_timelock": 5}],
            )
        [tx2] = action_scope.side_effects.transactions
        clawback_coin_id_2 = tx2.additions[0].name()
        assert tx2.spend_bundle is not None

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount2,
                            "<=#spendable_balance": -1 * tx_amount2,
                            "<=#max_send_amount": -1 * tx_amount2,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount2,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        # Check merkle coins
        await time_out_assert(20, wsm_1.coin_store.count_small_unspent, 2, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 1, 1000, CoinType.CLAWBACK)
        # clawback merkle coin
        resp = await api_1.spend_clawback_coins({"coin_ids": [clawback_coin_id_1.hex()], "fee": 0})
        assert resp["success"]
        assert len(resp["transaction_ids"]) == 1
        resp = await api_1.spend_clawback_coins({"coin_ids": [clawback_coin_id_2.hex()], "fee": 0})
        assert resp["success"]
        assert len(resp["transaction_ids"]) == 1

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": tx_amount + tx_amount2,
                            "pending_change": tx_amount + tx_amount2,
                            "pending_coin_removal_count": 2,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount + tx_amount2,
                            "max_send_amount": tx_amount + tx_amount2,
                            "spendable_balance": tx_amount + tx_amount2,
                            "pending_change": -1 * (tx_amount + tx_amount2),
                            "unspent_coin_count": 2,
                            "pending_coin_removal_count": -2,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={},
                ),
            ]
        )

        await time_out_assert(20, wsm_1.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)
        await time_out_assert(20, wsm_2.coin_store.count_small_unspent, 0, 1000, CoinType.CLAWBACK)

        before_txs: dict[str, dict[TransactionType, int]] = {"sender": {}, "recipient": {}}
        before_txs["sender"][
            TransactionType.INCOMING_CLAWBACK_SEND
        ] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_CLAWBACK_SEND])
        )
        before_txs["sender"][TransactionType.OUTGOING_CLAWBACK] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.OUTGOING_CLAWBACK])
        )
        before_txs["sender"][TransactionType.OUTGOING_TX] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.OUTGOING_TX])
        )
        before_txs["sender"][TransactionType.INCOMING_TX] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_TX])
        )
        before_txs["sender"][TransactionType.COINBASE_REWARD] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.COINBASE_REWARD])
        )
        before_txs["recipient"][
            TransactionType.INCOMING_CLAWBACK_RECEIVE
        ] = await wsm_2.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_CLAWBACK_RECEIVE])
        )
        # Resync start
        env_1.node._close()
        await env_1.node._await_closed()
        env_2.node._close()
        await env_2.node._await_closed()
        env_1.node.config["database_path"] = "wallet/db/blockchain_wallet_v2_test1_CHALLENGE_KEY.sqlite"
        env_2.node.config["database_path"] = "wallet/db/blockchain_wallet_v2_test2_CHALLENGE_KEY.sqlite"

        # use second node to start the same wallet, reusing config and db
        await env_1.node._start()
        await env_1.peer_server.start_client(PeerInfo(self_hostname, full_node_api.full_node.server.get_port()), None)
        await env_2.node._start()
        await env_2.peer_server.start_client(PeerInfo(self_hostname, full_node_api.full_node.server.get_port()), None)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(),
                WalletStateTransition(),
            ]
        )

        wsm_1 = env_1.node.wallet_state_manager
        wsm_2 = env_2.node.wallet_state_manager

        after_txs: dict[str, dict[TransactionType, int]] = {"sender": {}, "recipient": {}}
        after_txs["sender"][
            TransactionType.INCOMING_CLAWBACK_SEND
        ] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_CLAWBACK_SEND])
        )
        after_txs["sender"][TransactionType.OUTGOING_CLAWBACK] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.OUTGOING_CLAWBACK])
        )
        after_txs["sender"][TransactionType.OUTGOING_TX] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.OUTGOING_TX])
        )
        after_txs["sender"][TransactionType.INCOMING_TX] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_TX])
        )
        after_txs["sender"][TransactionType.COINBASE_REWARD] = await wsm_1.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.COINBASE_REWARD])
        )
        after_txs["recipient"][
            TransactionType.INCOMING_CLAWBACK_RECEIVE
        ] = await wsm_2.tx_store.get_transaction_count_for_wallet(
            1, type_filter=TransactionTypeFilter.include([TransactionType.INCOMING_CLAWBACK_RECEIVE])
        )
        # Check clawback
        clawback_tx_1 = await wsm_1.tx_store.get_transaction_record(clawback_coin_id_1)
        clawback_tx_2 = await wsm_1.tx_store.get_transaction_record(clawback_coin_id_2)
        assert clawback_tx_1 is not None
        assert clawback_tx_1.confirmed
        assert clawback_tx_2 is not None
        assert clawback_tx_2.confirmed
        outgoing_clawback_txs = await wsm_1.tx_store.get_transactions_between(
            1, 0, 100, type_filter=TransactionTypeFilter.include([TransactionType.OUTGOING_CLAWBACK])
        )
        assert len(outgoing_clawback_txs) == 2
        assert outgoing_clawback_txs[0].confirmed
        assert outgoing_clawback_txs[1].confirmed

        # transactions should be the same

        assert (
            before_txs["sender"][TransactionType.OUTGOING_CLAWBACK]
            == after_txs["sender"][TransactionType.OUTGOING_CLAWBACK]
        )
        assert before_txs["sender"] == after_txs["sender"]
        assert before_txs["recipient"] == after_txs["recipient"]

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [3]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_coinbase_reorg(self, wallet_environments: WalletTestFramework) -> None:
        full_node_api = wallet_environments.full_node
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        peak = full_node_api.full_node.blockchain.get_peak()
        assert peak is not None
        permanent_height = peak.height  # The height of the blocks we will not reorg

        extra_blocks = 2
        await full_node_api.farm_blocks_to_wallet(count=extra_blocks, wallet=wallet)
        await full_node_api.wait_for_wallet_synced(wallet_node=env.node, timeout=5)
        await env.change_balances(
            {
                1: {
                    "confirmed_wallet_balance": 2_000_000_000_000 * extra_blocks,
                    "unconfirmed_wallet_balance": 2_000_000_000_000 * extra_blocks,
                    "max_send_amount": 2_000_000_000_000 * extra_blocks,
                    "spendable_balance": 2_000_000_000_000 * extra_blocks,
                    "unspent_coin_count": 4,
                }
            }
        )

        await full_node_api.reorg_from_index_to_new_index(
            ReorgProtocol(
                uint32(permanent_height), uint32(permanent_height + extra_blocks + 6), bytes32(32 * b"0"), None
            )
        )

        await full_node_api.wait_for_wallet_synced(wallet_node=env.node, timeout=5)

        await env.change_balances(
            {
                1: {
                    "confirmed_wallet_balance": -2_000_000_000_000 * extra_blocks,
                    "unconfirmed_wallet_balance": -2_000_000_000_000 * extra_blocks,
                    "max_send_amount": -2_000_000_000_000 * extra_blocks,
                    "spendable_balance": -2_000_000_000_000 * extra_blocks,
                    "unspent_coin_count": -4,
                }
            }
        )

    @pytest.mark.parametrize("trusted", [True, False])
    @pytest.mark.anyio
    async def test_wallet_send_to_three_peers(
        self,
        three_sim_two_wallets: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
        trusted: bool,
        self_hostname: str,
    ) -> None:
        num_blocks = 10
        full_nodes, wallets, _ = three_sim_two_wallets

        wallet_0, wallet_server_0 = wallets[0]

        full_node_api_0 = full_nodes[0]
        full_node_api_1 = full_nodes[1]
        full_node_api_2 = full_nodes[2]

        full_node_0 = full_node_api_0.full_node
        full_node_1 = full_node_api_1.full_node
        full_node_2 = full_node_api_2.full_node

        server_0 = full_node_0.server
        server_1 = full_node_1.server
        server_2 = full_node_2.server

        if trusted:
            wallet_0.config["trusted_peers"] = {
                server_0.node_id.hex(): server_0.node_id.hex(),
                server_1.node_id.hex(): server_1.node_id.hex(),
                server_2.node_id.hex(): server_2.node_id.hex(),
            }

        else:
            wallet_0.config["trusted_peers"] = {}

        # wallet0 <-> sever0
        await wallet_server_0.start_client(PeerInfo(self_hostname, server_0.get_port()), None)

        await full_node_api_0.farm_blocks_to_wallet(count=num_blocks, wallet=wallet_0.wallet_state_manager.main_wallet)

        all_blocks = await full_node_api_0.get_all_full_blocks()

        for block in all_blocks:
            await full_node_1.add_block(block)
            await full_node_2.add_block(block)

        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_0.wallet_state_manager.main_wallet.generate_signed_transaction(
                uint64(10),
                bytes32(32 * b"0"),
                action_scope,
                uint64(0),
            )
        await full_node_api_0.wait_transaction_records_entered_mempool(records=action_scope.side_effects.transactions)

        # wallet0 <-> sever1
        await wallet_server_0.start_client(PeerInfo(self_hostname, server_1.get_port()), wallet_0.on_connect)
        await full_node_api_1.wait_transaction_records_entered_mempool(records=action_scope.side_effects.transactions)

        # wallet0 <-> sever2
        await wallet_server_0.start_client(PeerInfo(self_hostname, server_2.get_port()), wallet_0.on_connect)
        await full_node_api_2.wait_transaction_records_entered_mempool(records=action_scope.side_effects.transactions)

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_make_transaction_hop(self, wallet_environments: WalletTestFramework) -> None:
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet

        tx_amount = 10
        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_0.generate_signed_transaction(
                uint64(tx_amount),
                await wallet_1.get_puzzle_hash(False),
                action_scope,
                uint64(0),
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                        }
                    },
                ),
            ]
        )

        tx_amount = 5
        async with wallet_1.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_1.generate_signed_transaction(
                uint64(tx_amount), await wallet_0.get_puzzle_hash(False), action_scope, uint64(0)
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
            ]
        )

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_make_transaction_with_fee(self, wallet_environments: WalletTestFramework) -> None:
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet

        tx_amount = 1_750_000_000_000  # ensures we grab both coins
        tx_fee = 10
        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_0.generate_signed_transaction(
                uint64(tx_amount),
                await wallet_1.get_new_puzzlehash(),
                action_scope,
                uint64(tx_fee),
            )
        [tx] = action_scope.side_effects.transactions
        assert tx.spend_bundle is not None

        fees = estimate_fees(tx.spend_bundle)
        assert fees == tx_fee

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount - tx_fee,
                            "<=#spendable_balance": -1 * tx_amount - tx_fee,
                            "<=#max_send_amount": -1 * tx_amount - tx_fee,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 2,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount - tx_fee,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -2,
                            "unspent_coin_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                        }
                    },
                ),
            ]
        )

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [1, 1]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_make_transaction_with_memo(self, wallet_environments: WalletTestFramework) -> None:
        env_0 = wallet_environments.environments[0]
        env_1 = wallet_environments.environments[1]
        wallet_0 = env_0.xck_wallet
        wallet_1 = env_1.xck_wallet

        tx_amount = 1_750_000_000_000  # ensures we grab both coins
        tx_fee = 10
        ph_2 = await wallet_1.get_new_puzzlehash()
        async with wallet_0.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet_0.generate_signed_transaction(
                uint64(tx_amount), ph_2, action_scope, uint64(tx_fee), memos=[ph_2]
            )
        [tx] = action_scope.side_effects.transactions
        assert tx.spend_bundle is not None

        fees = estimate_fees(tx.spend_bundle)
        assert fees == tx_fee

        memos = await env_0.rpc_client.get_transaction_memo(GetTransactionMemo(transaction_id=tx.name))
        assert len(memos.coins_with_memos) == 1
        assert memos.coins_with_memos[0].memos[0] == ph_2

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount - tx_fee,
                            "<=#spendable_balance": -1 * tx_amount - tx_fee,
                            "<=#max_send_amount": -1 * tx_amount - tx_fee,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 2,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount - tx_fee,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -2,
                            "unspent_coin_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                        }
                    },
                ),
            ]
        )

        tx_id = None
        for coin in tx.additions:
            if coin.amount == tx_amount:
                tx_id = coin.name()
        assert tx_id is not None
        memos = await env_1.rpc_client.get_transaction_memo(GetTransactionMemo(transaction_id=tx_id))
        assert len(memos.coins_with_memos) == 1
        assert memos.coins_with_memos[0].memos[0] == ph_2
        # test json serialization
        assert memos.to_json_dict() == {
            tx_id.hex(): {memos.coins_with_memos[0].coin_id.hex(): [memos.coins_with_memos[0].memos[0].hex()]}
        }

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [1], "trusted": True, "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_create_hit_max_send_amount(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        ph = await wallet.get_puzzle_hash(False)
        primaries = [Payment(ph, uint64(1000000000 + i)) for i in range(int(wallet.max_send_quantity) + 1)]
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(uint64(1), ph, action_scope, uint64(0), primaries=primaries)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            # tx sent to ourselves
                            "unconfirmed_wallet_balance": 0,
                            "<=#spendable_balance": 0,
                            "<=#max_send_amount": 0,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": 0,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                            "unspent_coin_count": len(primaries) + 1,
                        }
                    },
                ),
            ]
        )

        max_sent_amount = await wallet.get_max_send_amount()
        assert max_sent_amount < (await wallet.get_spendable_balance())

        # 1) Generate transaction that is under the limit
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=False) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(max_sent_amount - 1),
                ph,
                action_scope,
                uint64(0),
            )

        assert action_scope.side_effects.transactions[0].amount == uint64(max_sent_amount - 1)

        # 2) Generate transaction that is equal to limit
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=False) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(max_sent_amount),
                ph,
                action_scope,
                uint64(0),
            )

        assert action_scope.side_effects.transactions[0].amount == uint64(max_sent_amount)

        # 3) Generate transaction that is greater than limit
        with pytest.raises(
            ValueError,
            match=f"Transaction for {max_sent_amount + 1} is greater than max spendable balance in a block of "
            f"{max_sent_amount}. There may be other transactions pending or our minimum coin amount is too high.",
        ):
            async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=False) as action_scope:
                await wallet.generate_signed_transaction(
                    uint64(max_sent_amount + 1),
                    ph,
                    action_scope,
                    uint64(0),
                )

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [2], "trusted": True, "reuse_puzhash": True}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_prevent_fee_theft(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        tx_amount = 1_750_000_000_000
        tx_fee = 2_000_000_000_000
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=False) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount),
                bytes32.zeros,
                action_scope,
                uint64(tx_fee),
            )
        [tx] = action_scope.side_effects.transactions
        assert tx.spend_bundle is not None

        stolen_cs: Optional[CoinSpend] = None
        # extract coin_spend from generated spend_bundle
        for cs in tx.spend_bundle.coin_spends:
            if compute_additions(cs) == []:
                stolen_cs = cs

        assert stolen_cs is not None

        # get a legit signature
        stolen_sb, _ = await wallet.wallet_state_manager.sign_bundle([stolen_cs])
        name = stolen_sb.name()
        stolen_tx = TransactionRecord(
            confirmed_at_height=uint32(0),
            created_at_time=uint64(0),
            to_puzzle_hash=bytes32(32 * b"0"),
            amount=uint64(0),
            fee_amount=uint64(0),
            confirmed=False,
            sent=uint32(0),
            spend_bundle=stolen_sb,
            additions=[],
            removals=[],
            wallet_id=wallet.id(),
            sent_to=[],
            trade_id=None,
            type=uint32(TransactionType.OUTGOING_TX.value),
            name=name,
            memos=[],
            valid_times=ConditionValidTimes(),
        )
        [stolen_tx] = await wallet.wallet_state_manager.add_pending_transactions([stolen_tx])

        async def transaction_has_failed(tx_id: bytes32) -> bool:
            tx = await wallet.wallet_state_manager.tx_store.get_transaction_record(tx_id)
            assert tx is not None
            return any(error_str == Err.ASSERT_ANNOUNCE_CONSUMED_FAILED.name for _, _, error_str in tx.sent_to)

        await time_out_assert(10, transaction_has_failed, True, stolen_tx.name)

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 2, "blocks_needed": [4, 1]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_tx_reorg(self, wallet_environments: WalletTestFramework) -> None:
        full_node_api = wallet_environments.full_node
        env = wallet_environments.environments[0]
        env_2 = wallet_environments.environments[1]
        wsm = env.wallet_state_manager
        wallet = env.xck_wallet
        wallet_2 = env_2.xck_wallet

        # Ensure that we use a coin that we will not reorg out
        tx_amount = 1000
        async with wallet.wallet_state_manager.new_action_scope(
            wallet_environments.tx_config, push=False
        ) as action_scope:
            coins = await wallet.select_coins(amount=uint64(tx_amount), action_scope=action_scope)
        coin = next(iter(coins))

        reorg_height = full_node_api.full_node.blockchain.get_peak_height()
        assert reorg_height is not None
        await full_node_api.farm_blocks_to_puzzlehash(count=3)

        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                uint64(tx_amount), await wallet_2.get_puzzle_hash(False), action_scope, coins={coin}
            )

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "<=#spendable_balance": -1 * tx_amount,
                            "<=#max_send_amount": -1 * tx_amount,
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": 1,  # any amount increase
                            ">=#max_send_amount": 1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={},
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                        }
                    },
                ),
            ]
        )

        peak = full_node_api.full_node.blockchain.get_peak()
        assert peak is not None
        peak_height = peak.height
        assert peak_height is not None

        target_height_after_reorg = peak_height + 3
        # Perform a reorg, which will revert the transaction in the full node and wallet, and cause wallet to resubmit
        await full_node_api.reorg_from_index_to_new_index(
            ReorgProtocol(uint32(reorg_height - 1), uint32(target_height_after_reorg), bytes32(32 * b"0"), None)
        )

        await time_out_assert(20, full_node_api.full_node.blockchain.get_peak_height, target_height_after_reorg)

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": 0,
                            "<=#spendable_balance": -1,  # any amount decrease
                            "<=#max_send_amount": -1,  # any amount decrease
                            ">=#pending_change": 1,  # any amount increase
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            ">=#spendable_balance": -1,  # any amount increase
                            ">=#max_send_amount": -1,  # any amount increase
                            "<=#pending_change": -1,  # any amount decrease
                            "pending_coin_removal_count": -1,
                        }
                    },
                ),
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * tx_amount,
                            "unconfirmed_wallet_balance": -1 * tx_amount,
                            "spendable_balance": -1 * tx_amount,
                            "max_send_amount": -1 * tx_amount,
                            "unspent_coin_count": -1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": tx_amount,
                            "unconfirmed_wallet_balance": tx_amount,
                            "spendable_balance": tx_amount,
                            "max_send_amount": tx_amount,
                            "unspent_coin_count": 1,
                        }
                    },
                ),
            ]
        )

        unconfirmed = await wsm.tx_store.get_unconfirmed_for_wallet(int(wallet.id()))
        assert len(unconfirmed) == 0
        [tx] = action_scope.side_effects.transactions
        tx_record = await wsm.tx_store.get_transaction_record(tx.name)
        assert tx_record is not None
        removed = tx_record.removals[0]
        added = tx_record.additions[0]
        added_1 = tx_record.additions[1]
        wallet_coin_record_rem = await wsm.coin_store.get_coin_record(removed.name())
        assert wallet_coin_record_rem is not None
        assert wallet_coin_record_rem.spent

        coin_record_full_node = await full_node_api.full_node.coin_store.get_coin_record(removed.name())
        assert coin_record_full_node is not None
        assert coin_record_full_node.spent
        add_1_coin_record_full_node = await full_node_api.full_node.coin_store.get_coin_record(added.name())
        assert add_1_coin_record_full_node is not None
        assert add_1_coin_record_full_node.confirmed_block_index > 0
        add_2_coin_record_full_node = await full_node_api.full_node.coin_store.get_coin_record(added_1.name())
        assert add_2_coin_record_full_node is not None
        assert add_2_coin_record_full_node.confirmed_block_index > 0

    @pytest.mark.parametrize(
        "wallet_environments",
        [
            {
                "num_environments": 1,
                "blocks_needed": [1],
                "trusted": True,
                "reuse_puzhash": False,
                "config_overrides": {"initial_num_public_keys": 100},
            }
        ],
        indirect=True,
    )
    @pytest.mark.anyio
    async def test_address_sliding_window(self, wallet_environments: WalletTestFramework) -> None:
        full_node_api = wallet_environments.full_node
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        peak = full_node_api.full_node.blockchain.get_peak_height()
        assert peak is not None

        puzzle_hashes = []
        for i in range(211):
            pubkey = master_sk_to_wallet_sk(wallet.wallet_state_manager.get_master_private_key(), uint32(i)).get_g1()
            puzzle: Program = wallet.puzzle_for_pk(pubkey)
            puzzle_hash: bytes32 = puzzle.get_tree_hash()
            puzzle_hashes.append(puzzle_hash)

        await full_node_api.farm_blocks_to_puzzlehash(count=1, farm_to=puzzle_hashes[0])
        await full_node_api.farm_blocks_to_puzzlehash(count=1, farm_to=puzzle_hashes[210])
        await full_node_api.farm_blocks_to_puzzlehash(
            count=1,
            farm_to=puzzle_hashes[114],
            guarantee_transaction_blocks=True,
        )

        await full_node_api.wait_for_wallet_synced(env.node, peak_height=uint32(peak + 3))
        await env.change_balances(
            {
                1: {
                    "confirmed_wallet_balance": 2_000_000_000_000,
                    "unconfirmed_wallet_balance": 2_000_000_000_000,
                    "spendable_balance": 2_000_000_000_000,
                    "max_send_amount": 2_000_000_000_000,
                    "unspent_coin_count": 2,
                }
            }
        )

        await full_node_api.farm_blocks_to_puzzlehash(
            count=1,
            farm_to=puzzle_hashes[50],
            guarantee_transaction_blocks=True,
        )
        await full_node_api.farm_blocks_to_puzzlehash(
            count=1,
            guarantee_transaction_blocks=True,
        )

        await full_node_api.wait_for_wallet_synced(env.node, peak_height=uint32(peak + 5))
        await env.change_balances(
            {
                1: {
                    "confirmed_wallet_balance": 6_000_000_000_000,
                    "unconfirmed_wallet_balance": 6_000_000_000_000,
                    "spendable_balance": 6_000_000_000_000,
                    "max_send_amount": 6_000_000_000_000,
                    "unspent_coin_count": 6,
                }
            }
        )

        await full_node_api.farm_blocks_to_puzzlehash(count=1, farm_to=puzzle_hashes[113])
        await full_node_api.farm_blocks_to_puzzlehash(
            count=1,
            farm_to=puzzle_hashes[209],
            guarantee_transaction_blocks=True,
        )
        await full_node_api.farm_blocks_to_puzzlehash(count=1, guarantee_transaction_blocks=True)

        await full_node_api.wait_for_wallet_synced(env.node, peak_height=uint32(peak + 8))
        await env.change_balances(
            {
                1: {
                    "confirmed_wallet_balance": 4_000_000_000_000,
                    "unconfirmed_wallet_balance": 4_000_000_000_000,
                    "spendable_balance": 4_000_000_000_000,
                    "max_send_amount": 4_000_000_000_000,
                    "unspent_coin_count": 4,
                }
            }
        )

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [1]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_sign_message(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        api_0 = env.rpc_api

        # Test general string
        message = "Hello World"
        ph = await env.xck_wallet.get_puzzle_hash(False)
        response = await api_0.sign_message_by_address({"address": encode_puzzle_hash(ph, "xck"), "message": message})
        puzzle: Program = Program.to((CHIP_0002_SIGN_MESSAGE_PREFIX, message))

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            puzzle.get_tree_hash(),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )
        # Test hex string
        message = "0123456789ABCDEF"
        response = await api_0.sign_message_by_address(
            {"address": encode_puzzle_hash(ph, "xck"), "message": message, "is_hex": True}
        )
        puzzle = Program.to((CHIP_0002_SIGN_MESSAGE_PREFIX, bytes.fromhex(message)))

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            puzzle.get_tree_hash(),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )
        # Test informal input
        message = "0123456789ABCDEF"
        response = await api_0.sign_message_by_address(
            {"address": encode_puzzle_hash(ph, "xck"), "message": message, "is_hex": "true", "safe_mode": "true"}
        )
        puzzle = Program.to((CHIP_0002_SIGN_MESSAGE_PREFIX, bytes.fromhex(message)))

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            puzzle.get_tree_hash(),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )
        # Test BLS sign string
        message = "Hello World"
        response = await api_0.sign_message_by_address(
            {"address": encode_puzzle_hash(ph, "xck"), "message": message, "is_hex": False, "safe_mode": False}
        )

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            bytes(message, "utf-8"),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )
        # Test BLS sign hex
        message = "0123456789ABCDEF"
        response = await api_0.sign_message_by_address(
            {"address": encode_puzzle_hash(ph, "xck"), "message": message, "is_hex": True, "safe_mode": False}
        )

        assert AugSchemeMPL.verify(
            G1Element.from_bytes(bytes.fromhex(response["pubkey"])),
            bytes.fromhex(message),
            G2Element.from_bytes(bytes.fromhex(response["signature"])),
        )

    @pytest.mark.parametrize(
        "wallet_environments",
        [{"num_environments": 1, "blocks_needed": [2]}],
        indirect=True,
    )
    @pytest.mark.limit_consensus_modes(reason="irrelevant")
    @pytest.mark.anyio
    async def test_wallet_transaction_options(self, wallet_environments: WalletTestFramework) -> None:
        env = wallet_environments.environments[0]
        wallet = env.xck_wallet

        AMOUNT_TO_SEND = 4000000000000
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            coins = await wallet.select_coins(uint64(AMOUNT_TO_SEND), action_scope)
            coin_list = list(coins)
            await wallet.generate_signed_transaction(
                uint64(AMOUNT_TO_SEND),
                bytes32.zeros,
                action_scope,
                uint64(0),
                coins=coins,
                origin_id=coin_list[2].name(),
            )
        [tx] = action_scope.side_effects.transactions
        assert tx.spend_bundle is not None
        paid_coin = next(coin for coin in tx.spend_bundle.additions() if coin.amount == AMOUNT_TO_SEND)
        assert paid_coin.parent_coin_info == coin_list[2].name()
        [tx] = await wallet.wallet_state_manager.add_pending_transactions([tx])

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * AMOUNT_TO_SEND,
                            "spendable_balance": -1 * AMOUNT_TO_SEND,  # used exact amount
                            "max_send_amount": -1 * AMOUNT_TO_SEND,  # used exact amount
                            "pending_change": 0,  # used exact amount
                            "pending_coin_removal_count": len(coins),
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * AMOUNT_TO_SEND,
                            "spendable_balance": 0,  # used exact amount
                            "max_send_amount": 0,  # used exact amount
                            "pending_change": 0,  # used exact amount
                            "unspent_coin_count": -len(coins),
                            "pending_coin_removal_count": -len(coins),
                        }
                    },
                )
            ]
        )


def test_get_wallet_db_path_v2_r1() -> None:
    root_path: Path = Path("/x/y/z/.chik/mainnet").resolve()
    config: dict[str, Any] = {
        "database_path": "wallet/db/blockchain_wallet_v2_r1_CHALLENGE_KEY.sqlite",
        "selected_network": "mainnet",
    }
    fingerprint: str = "1234567890"
    wallet_db_path: Path = get_wallet_db_path(root_path, config, fingerprint)

    assert wallet_db_path == root_path.joinpath("wallet/db/blockchain_wallet_v2_r1_mainnet_1234567890.sqlite")


def test_get_wallet_db_path_v2() -> None:
    root_path: Path = Path("/x/y/z/.chik/mainnet").resolve()
    config: dict[str, Any] = {
        "database_path": "wallet/db/blockchain_wallet_v2_CHALLENGE_KEY.sqlite",
        "selected_network": "mainnet",
    }
    fingerprint: str = "1234567890"
    wallet_db_path: Path = get_wallet_db_path(root_path, config, fingerprint)

    assert wallet_db_path == root_path.joinpath("wallet/db/blockchain_wallet_v2_r1_mainnet_1234567890.sqlite")


def test_get_wallet_db_path_v1() -> None:
    root_path: Path = Path("/x/y/z/.chik/mainnet").resolve()
    config: dict[str, Any] = {
        "database_path": "wallet/db/blockchain_wallet_v1_CHALLENGE_KEY.sqlite",
        "selected_network": "mainnet",
    }
    fingerprint: str = "1234567890"
    wallet_db_path: Path = get_wallet_db_path(root_path, config, fingerprint)

    assert wallet_db_path == root_path.joinpath("wallet/db/blockchain_wallet_v2_r1_mainnet_1234567890.sqlite")


def test_get_wallet_db_path_testnet() -> None:
    root_path: Path = Path("/x/y/z/.chik/testnet").resolve()
    config: dict[str, Any] = {
        "database_path": "wallet/db/blockchain_wallet_v2_CHALLENGE_KEY.sqlite",
        "selected_network": "testnet",
    }
    fingerprint: str = "1234567890"
    wallet_db_path: Path = get_wallet_db_path(root_path, config, fingerprint)

    assert wallet_db_path == root_path.joinpath("wallet/db/blockchain_wallet_v2_r1_testnet_1234567890.sqlite")


@pytest.mark.anyio
async def test_wallet_has_no_server(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    _full_nodes, wallets, _bt = simulator_and_wallet
    _wallet_node, wallet_server = wallets[0]

    assert wallet_server.webserver is None



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional, TypeVar

from chik.protocols.wallet_protocol import CoinState
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend
from chik.util.ints import uint16
from chik.util.streamable import Streamable, streamable
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile

log = logging.getLogger(__name__)
SINGLETON_TOP_LAYER_MOD = load_klvm_maybe_recompile("singleton_top_layer_v1_1.clsp")
NFT_MOD = load_klvm_maybe_recompile("nft_state_layer.clsp", package_or_requirement="chik.wallet.nft_wallet.puzzles")
NFT_OWNERSHIP_LAYER = load_klvm_maybe_recompile(
    "nft_ownership_layer.clsp", package_or_requirement="chik.wallet.nft_wallet.puzzles"
)

_T_UncurriedNFT = TypeVar("_T_UncurriedNFT", bound="UncurriedNFT")


@streamable
@dataclass(frozen=True)
class UncurriedNFT(Streamable):
    """
    A simple solution for uncurry NFT puzzle.
    Initial the class with a full NFT puzzle, it will do a deep uncurry.
    This is the only place you need to change after modified the Chiklisp curried parameters.
    """

    nft_mod_hash: bytes32
    """NFT module hash"""

    nft_state_layer: Program
    """NFT state layer puzzle"""

    singleton_struct: Program
    """
    Singleton struct
    [singleton_mod_hash, singleton_launcher_id, launcher_puzhash]
    """
    singleton_mod_hash: Program
    singleton_launcher_id: bytes32
    launcher_puzhash: Program

    metadata_updater_hash: Program
    """Metadata updater puzzle hash"""

    metadata: Program
    """
    NFT metadata
    [("u", data_uris), ("h", data_hash)]
    """
    data_uris: Program
    data_hash: Program
    meta_uris: Program
    meta_hash: Program
    license_uris: Program
    license_hash: Program
    edition_number: Program
    edition_total: Program

    inner_puzzle: Program
    """NFT state layer inner puzzle"""

    p2_puzzle: Program
    """p2 puzzle of the owner, either for ownership layer or standard"""

    # ownership layer fields
    owner_did: Optional[bytes32]
    """Owner's DID"""

    supports_did: bool
    """If the inner puzzle support the DID"""

    nft_inner_puzzle_hash: Optional[bytes32]
    """Puzzle hash of the ownership layer inner puzzle """

    transfer_program: Optional[Program]
    """Puzzle hash of the transfer program"""

    transfer_program_curry_params: Optional[Program]
    """
    Curried parameters of the transfer program
    [royalty_address, trade_price_percentage, settlement_mod_hash, cat_mod_hash]
    """
    royalty_address: Optional[bytes32]
    trade_price_percentage: Optional[uint16]

    @classmethod
    def uncurry(cls: type[_T_UncurriedNFT], mod: Program, curried_args: Program) -> Optional[_T_UncurriedNFT]:
        """
        Try to uncurry a NFT puzzle
        :param cls UncurriedNFT class
        :param mod: uncurried Puzzle program
        :param uncurried_args: uncurried arguments to program
        :return Uncurried NFT
        """
        if mod != SINGLETON_TOP_LAYER_MOD:
            log.debug("Cannot uncurry NFT puzzle, failed on singleton top layer: Mod %s", mod)
            return None
        try:
            (singleton_struct, nft_state_layer) = curried_args.as_iter()
            singleton_mod_hash = singleton_struct.first()
            singleton_launcher_id = singleton_struct.rest().first()
            launcher_puzhash = singleton_struct.rest().rest()
        except ValueError as e:
            log.debug("Cannot uncurry singleton top layer: Args %s error: %s", curried_args, e)
            return None

        mod, curried_args = curried_args.rest().first().uncurry()
        if mod != NFT_MOD:
            log.debug("Cannot uncurry NFT puzzle, failed on NFT state layer: Mod %s", mod)
            return None
        try:
            # Set nft parameters
            nft_mod_hash, metadata, metadata_updater_hash, inner_puzzle = curried_args.as_iter()
            data_uris = Program.to([])
            data_hash = Program.to(0)
            meta_uris = Program.to([])
            meta_hash = Program.to(0)
            license_uris = Program.to([])
            license_hash = Program.to(0)
            edition_number = Program.to(1)
            edition_total = Program.to(1)
            # Set metadata
            for kv_pair in metadata.as_iter():
                if kv_pair.first().as_atom() == b"u":
                    data_uris = kv_pair.rest()
                if kv_pair.first().as_atom() == b"h":
                    data_hash = kv_pair.rest()
                if kv_pair.first().as_atom() == b"mu":
                    meta_uris = kv_pair.rest()
                if kv_pair.first().as_atom() == b"mh":
                    meta_hash = kv_pair.rest()
                if kv_pair.first().as_atom() == b"lu":
                    license_uris = kv_pair.rest()
                if kv_pair.first().as_atom() == b"lh":
                    license_hash = kv_pair.rest()
                if kv_pair.first().as_atom() == b"sn":
                    edition_number = kv_pair.rest()
                if kv_pair.first().as_atom() == b"st":
                    edition_total = kv_pair.rest()
            current_did: Optional[bytes32] = None
            transfer_program = None
            transfer_program_args = None
            royalty_address: Optional[bytes32] = None
            royalty_percentage: Optional[uint16] = None
            nft_inner_puzzle_mod = None
            mod, ol_args = inner_puzzle.uncurry()
            supports_did = False
            if mod == NFT_OWNERSHIP_LAYER:
                supports_did = True
                log.debug("Parsing ownership layer")
                _, current_did_p, transfer_program, p2_puzzle = ol_args.as_iter()
                _, transfer_program_args = transfer_program.uncurry()
                _, royalty_address_p, royalty_percentage_p = transfer_program_args.as_iter()
                royalty_percentage = uint16(royalty_percentage_p.as_int())
                royalty_address = bytes32(royalty_address_p.as_atom())
                current_did_atom = current_did_p.as_atom()
                if current_did_atom == b"":
                    # For unassigned NFT, set owner DID to None
                    current_did = None
                else:
                    current_did = bytes32(current_did_atom)
            else:
                log.debug("Creating a standard NFT puzzle")
                p2_puzzle = inner_puzzle
        except Exception as e:
            log.debug("Cannot uncurry NFT state layer: Args %s Error: %s", curried_args, e)
            return None

        return cls(
            nft_mod_hash=bytes32(nft_mod_hash.as_atom()),
            nft_state_layer=nft_state_layer,
            singleton_struct=singleton_struct,
            singleton_mod_hash=singleton_mod_hash,
            singleton_launcher_id=bytes32(singleton_launcher_id.as_atom()),
            launcher_puzhash=launcher_puzhash,
            metadata=metadata,
            data_uris=data_uris,
            data_hash=data_hash,
            p2_puzzle=p2_puzzle,
            metadata_updater_hash=metadata_updater_hash,
            meta_uris=meta_uris,
            meta_hash=meta_hash,
            license_uris=license_uris,
            license_hash=license_hash,
            edition_number=edition_number,
            edition_total=edition_total,
            inner_puzzle=inner_puzzle,
            owner_did=current_did,
            supports_did=supports_did,
            transfer_program=transfer_program,
            transfer_program_curry_params=transfer_program_args,
            royalty_address=royalty_address,
            trade_price_percentage=royalty_percentage,
            nft_inner_puzzle_hash=nft_inner_puzzle_mod,
        )

    def get_innermost_solution(self, solution: Program) -> Program:
        state_layer_inner_solution: Program = solution.at("rrff")
        if self.supports_did:
            return state_layer_inner_solution.first()
        else:
            return state_layer_inner_solution


@streamable
@dataclass(frozen=True)
class NFTCoinData(Streamable):
    uncurried_nft: UncurriedNFT
    parent_coin_state: CoinState
    parent_coin_spend: CoinSpend



from __future__ import annotations

import ipaddress
from dataclasses import dataclass
from typing import Union

from chik.util.ints import uint16, uint64
from chik.util.ip_address import IPAddress
from chik.util.streamable import Streamable, streamable


@dataclass(frozen=True)
class UnresolvedPeerInfo:
    host: str
    port: uint16


# TODO, Replace unsafe_hash with frozen and drop the __init__ as soon as all PeerInfo call sites pass in an IPAddress.
@dataclass(unsafe_hash=True)
class PeerInfo:
    _ip: IPAddress
    _port: uint16

    # TODO, Drop this as soon as all call PeerInfo calls pass in an IPAddress
    def __init__(self, host: Union[IPAddress, str], port: int):
        self._ip = host if isinstance(host, IPAddress) else IPAddress.create(host)
        self._port = uint16(port)

    # Kept here for compatibility until all places where its used transitioned to IPAddress instead of str.
    @property
    def host(self) -> str:
        return str(self._ip)

    @property
    def ip(self) -> IPAddress:
        return self._ip

    @property
    def port(self) -> uint16:
        return self._port

    # Functions related to peer bucketing in new/tried tables.
    def get_key(self) -> bytes:
        if self.ip.is_v4:
            key = ipaddress.IPv6Address(int(ipaddress.IPv6Address("2002::")) | (int(self.ip) << 80)).packed
        else:
            key = self.ip.packed
        key += bytes([self.port // 0x100, self.port & 0x0FF])
        return key

    def get_group(self) -> bytes:
        # TODO: Port everything from Bitcoin.
        if self.ip.is_v4:
            return bytes([1]) + self.ip.packed[:2]
        else:
            return bytes([0]) + self.ip.packed[:4]


@streamable
@dataclass(frozen=True)
class TimestampedPeerInfo(Streamable):
    host: str
    port: uint16
    timestamp: uint64



from __future__ import annotations

import asyncio
import json
import pathlib
import time
from decimal import Decimal
from typing import Optional

from chik.cmds.cmds_util import CMDTXConfigLoader, get_wallet_client, transaction_status_msg, transaction_submitted_msg
from chik.cmds.param_types import CliAmount
from chik.cmds.units import units
from chik.cmds.wallet_funcs import get_mojo_per_unit, get_wallet_type
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import decode_puzzle_hash, encode_puzzle_hash
from chik.util.config import selected_network_address_prefix
from chik.util.ints import uint64
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.transaction_record import TransactionRecord
from chik.wallet.util.tx_config import DEFAULT_COIN_SELECTION_CONFIG
from chik.wallet.util.wallet_types import WalletType


async def add_dao_wallet(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    name: Optional[str],
    treasury_id: bytes32,
    filter_amount: uint64,
) -> None:
    print(f"Adding wallet for DAO: {treasury_id}")
    print("This may take awhile.")

    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.create_new_dao_wallet(
            mode="existing",
            tx_config=CMDTXConfigLoader(reuse_puzhash=True).to_tx_config(units["chik"], config, fingerprint),
            dao_rules=None,
            amount_of_cats=None,
            treasury_id=treasury_id,
            filter_amount=filter_amount,
            name=name,
        )

        print("Successfully created DAO Wallet")
        print(f"DAO Treasury ID: {res.treasury_id.hex()}")
        print(f"DAO Wallet ID: {res.wallet_id}")
        print(f"CAT Wallet ID: {res.cat_wallet_id}")
        print(f"DAOCAT Wallet ID: {res.dao_cat_wallet_id}")


async def create_dao_wallet(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    fee: uint64,
    fee_for_cat: uint64,
    name: Optional[str],
    proposal_timelock: uint64,
    soft_close: uint64,
    attendance_required: uint64,
    pass_percentage: uint64,
    self_destruct: uint64,
    oracle_delay: uint64,
    proposal_minimum: uint64,
    filter_amount: uint64,
    cat_amount: CliAmount,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    if proposal_minimum % 2 == 0:
        proposal_minimum = uint64(1 + proposal_minimum)
        print("Adding 1 mojo to proposal minimum amount")

    dao_rules = {
        "proposal_timelock": proposal_timelock,
        "soft_close_length": soft_close,
        "attendance_required": attendance_required,
        "pass_percentage": pass_percentage,
        "self_destruct_length": self_destruct,
        "oracle_spend_delay": oracle_delay,
        "proposal_minimum_amount": proposal_minimum,
    }

    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        conf_coins, _, _ = await wallet_client.get_spendable_coins(
            wallet_id=1, coin_selection_config=DEFAULT_COIN_SELECTION_CONFIG
        )
        if len(conf_coins) < 2:  # pragma: no cover
            raise ValueError("DAO creation requires at least 2 xck coins in your wallet.")
        res = await wallet_client.create_new_dao_wallet(
            mode="new",
            dao_rules=dao_rules,
            amount_of_cats=cat_amount.convert_amount(units["mojo"]),
            treasury_id=None,
            filter_amount=filter_amount,
            name=name,
            fee=fee,
            fee_for_cat=fee_for_cat,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        if push:
            print("Successfully created DAO Wallet")
        print(f"DAO Treasury ID: {res.treasury_id.hex()}")
        print(f"DAO Wallet ID: {res.wallet_id}")
        print(f"CAT Wallet ID: {res.cat_wallet_id}")
        print(f"DAOCAT Wallet ID: {res.dao_cat_wallet_id}")
        return res.transactions


async def get_treasury_id(root_path: pathlib.Path, wallet_rpc_port: Optional[int], fp: int, wallet_id: int) -> None:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, _, _):
        res = await wallet_client.dao_get_treasury_id(wallet_id=wallet_id)
        treasury_id = res["treasury_id"]
        print(f"Treasury ID: {treasury_id}")


async def get_rules(root_path: pathlib.Path, wallet_rpc_port: Optional[int], fp: int, wallet_id: int) -> None:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, _, _):
        res = await wallet_client.dao_get_rules(wallet_id=wallet_id)
        rules = res["rules"]
        for rule, val in rules.items():
            print(f"{rule}: {val}")


async def add_funds_to_treasury(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    funding_wallet_id: int,
    amount: CliAmount,
    fee: uint64,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        try:
            typ = await get_wallet_type(wallet_id=funding_wallet_id, wallet_client=wallet_client)
            mojo_per_unit = get_mojo_per_unit(typ)
        except LookupError:  # pragma: no cover
            print(f"Wallet id: {wallet_id} not found.")
            return []

        res = await wallet_client.dao_add_funds_to_treasury(
            wallet_id=wallet_id,
            funding_wallet_id=funding_wallet_id,
            amount=amount.convert_amount(mojo_per_unit),
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        if push:
            start = time.time()
            while time.time() - start < 10:
                await asyncio.sleep(0.1)
                tx = await wallet_client.get_transaction(res.tx_id)
                if len(tx.sent_to) > 0:
                    print(transaction_submitted_msg(tx))
                    print(transaction_status_msg(fingerprint, res.tx_id))
                    return res.transactions

        if push:
            print(f"Transaction not yet submitted to nodes. TX ID: {res.tx_id.hex()}")
        return res.transactions


async def get_treasury_balance(
    root_path: pathlib.Path, wallet_rpc_port: Optional[int], fp: int, wallet_id: int
) -> None:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, _, _):
        res = await wallet_client.dao_get_treasury_balance(wallet_id=wallet_id)
        balances = res["balances"]

        if not balances:
            print("The DAO treasury currently has no funds")
            return None

        xck_mojos = get_mojo_per_unit(WalletType.STANDARD_WALLET)
        cat_mojos = get_mojo_per_unit(WalletType.CAT)
        for asset_id, balance in balances.items():
            if asset_id == "xck":
                print(f"XCK: {balance / xck_mojos}")
            else:
                print(f"{asset_id}: {balance / cat_mojos}")


async def list_proposals(
    root_path: pathlib.Path, wallet_rpc_port: Optional[int], fp: int, wallet_id: int, include_closed: bool
) -> None:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, _, _):
        res = await wallet_client.dao_get_proposals(wallet_id=wallet_id, include_closed=include_closed)
        proposals = res["proposals"]
        soft_close_length = res["soft_close_length"]
        print("############################")
        for prop in proposals:
            print(f"Proposal ID: {prop['proposal_id']}")
            prop_status = "CLOSED" if prop["closed"] else "OPEN"
            print(f"Status: {prop_status}")
            print(f"Votes for: {prop['yes_votes']}")
            votes_against = prop["amount_voted"] - prop["yes_votes"]
            print(f"Votes against: {votes_against}")
            print("------------------------")
        print(f"Proposals have {soft_close_length} blocks of soft close time.")
        print("############################")


async def show_proposal(
    root_path: pathlib.Path, wallet_rpc_port: Optional[int], fp: int, wallet_id: int, proposal_id: str
) -> None:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, _, config):
        res = await wallet_client.dao_parse_proposal(wallet_id, proposal_id)
        pd = res["proposal_dictionary"]
        blocks_needed = pd["state"]["blocks_needed"]
        passed = pd["state"]["passed"]
        closable = pd["state"]["closable"]
        status = "CLOSED" if pd["state"]["closed"] else "OPEN"
        votes_needed = pd["state"]["total_votes_needed"]
        yes_needed = pd["state"]["yes_votes_needed"]

        ptype_val = pd["proposal_type"]
        if (ptype_val == "s") and ("mint_amount" in pd):
            ptype = "mint"
        elif ptype_val == "s":
            ptype = "spend"
        elif ptype_val == "u":
            ptype = "update"
        else:
            raise Exception(f"Unknown proposal type: {ptype_val!r}")

        print("")
        print(f"Details of Proposal: {proposal_id}")
        print("---------------------------")
        print("")
        print(f"Type: {ptype.upper()}")
        print(f"Status: {status}")
        print(f"Passed: {passed}")
        if not passed:
            print(f"Yes votes needed: {yes_needed}")

        if not pd["state"]["closed"]:
            print(f"Closable: {closable}")
            if not closable:
                print(f"Total votes needed: {votes_needed}")
                print(f"Blocks remaining: {blocks_needed}")

        prefix = selected_network_address_prefix(config)
        if ptype == "spend":
            xck_conds = pd["xck_conditions"]
            asset_conds = pd["asset_conditions"]
            print("")
            if xck_conds:
                print("Proposal XCK Conditions")
                for pmt in xck_conds:
                    puzzle_hash = encode_puzzle_hash(bytes32.from_hexstr(pmt["puzzle_hash"]), prefix)
                    amount = pmt["amount"]
                    print(f"Address: {puzzle_hash}\nAmount: {amount}\n")
            if asset_conds:
                print("Proposal asset Conditions")
                for cond in asset_conds:
                    asset_id = cond["asset_id"]
                    print(f"Asset ID: {asset_id}")
                    conds = cond["conditions"]
                    for pmt in conds:
                        puzzle_hash = encode_puzzle_hash(bytes32.from_hexstr(pmt["puzzle_hash"]), prefix)
                        amount = pmt["amount"]
                        print(f"Address: {puzzle_hash}\nAmount: {amount}\n")

        elif ptype == "update":
            print("")
            print("Proposed Rules:")
            for key, val in pd["dao_rules"].items():
                print(f"{key}: {val}")

        elif ptype == "mint":
            mint_amount = pd["mint_amount"]
            address = encode_puzzle_hash(bytes32.from_hexstr(pd["new_cat_puzhash"]), prefix)
            print("")
            print(f"Amount of CAT to mint: {mint_amount}")
            print(f"Address: {address}")


async def vote_on_proposal(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    proposal_id: str,
    vote_amount: uint64,
    is_yes_vote: bool,
    fee: uint64,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_vote_on_proposal(
            wallet_id=wallet_id,
            proposal_id=proposal_id,
            vote_amount=vote_amount,
            is_yes_vote=is_yes_vote,
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )
        if push:
            start = time.time()
            while time.time() - start < 10:
                await asyncio.sleep(0.1)
                tx = await wallet_client.get_transaction(res.tx_id)
                if len(tx.sent_to) > 0:
                    print(transaction_submitted_msg(tx))
                    print(transaction_status_msg(fingerprint, res.tx_id))
                    return res.transactions

        if push:
            print(f"Transaction not yet submitted to nodes. TX ID: {res.tx_id.hex()}")
        return res.transactions


async def close_proposal(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    fee: uint64,
    proposal_id: str,
    self_destruct: bool,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_close_proposal(
            wallet_id=wallet_id,
            proposal_id=proposal_id,
            fee=fee,
            self_destruct=self_destruct,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        if push:
            start = time.time()
            while time.time() - start < 10:
                await asyncio.sleep(0.1)
                tx = await wallet_client.get_transaction(res.tx_id)
                if len(tx.sent_to) > 0:
                    print(transaction_submitted_msg(tx))
                    print(transaction_status_msg(fingerprint, res.tx_id))
                    return res.transactions

        if push:
            print(f"Transaction not yet submitted to nodes. TX ID: {res.tx_id.hex()}")
        return res.transactions


async def lockup_coins(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    amount: CliAmount,
    fee: uint64,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    final_amount: uint64 = amount.convert_amount(units["cat"])
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_send_to_lockup(
            wallet_id=wallet_id,
            amount=final_amount,
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )
        if push:
            start = time.time()
            while time.time() - start < 10:
                await asyncio.sleep(0.1)
                tx = await wallet_client.get_transaction(res.tx_id)
                if len(tx.sent_to) > 0:
                    print(transaction_submitted_msg(tx))
                    print(transaction_status_msg(fingerprint, res.tx_id))
                    return res.transactions

        if push:
            print(f"Transaction not yet submitted to nodes. TX ID: {res.tx_id.hex()}")

        return res.transactions


async def release_coins(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    fee: uint64,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_free_coins_from_finished_proposals(
            wallet_id=wallet_id,
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )
        if push:
            start = time.time()
            while time.time() - start < 10:
                await asyncio.sleep(0.1)
                tx = await wallet_client.get_transaction(res.tx_id)
                if len(tx.sent_to) > 0:
                    print(transaction_submitted_msg(tx))
                    print(transaction_status_msg(fingerprint, res.tx_id))
                    return res.transactions

        if push:
            print(f"Transaction not yet submitted to nodes. TX ID: {res.tx_id.hex()}")
        return res.transactions


async def exit_lockup(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    fee: uint64,
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_exit_lockup(
            wallet_id=wallet_id,
            coins=[],
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        if push:
            start = time.time()
            while time.time() - start < 10:
                await asyncio.sleep(0.1)
                tx = await wallet_client.get_transaction(res.tx_id)
                if len(tx.sent_to) > 0:
                    print(transaction_submitted_msg(tx))
                    print(transaction_status_msg(fingerprint, res.tx_id))
                    return res.transactions

        if push:
            print(f"Transaction not yet submitted to nodes. TX ID: {res.tx_id.hex()}")
        return res.transactions


async def create_spend_proposal(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    fee: uint64,
    vote_amount: Optional[int],
    address: Optional[str],
    amount: Optional[str],
    asset_id: Optional[str],
    additions_file: Optional[str],
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    if additions_file is None and (address is None or amount is None):
        raise ValueError("Must include a json specification or an address / amount pair.")
    if additions_file:  # pragma: no cover
        with open(additions_file) as f:
            additions_dict = json.load(f)
        additions = []
        for addition in additions_dict:
            addition["puzzle_hash"] = decode_puzzle_hash(addition["address"]).hex()
            del addition["address"]
            additions.append(addition)
    else:
        additions = None
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        wallet_type = await get_wallet_type(wallet_id=wallet_id, wallet_client=wallet_client)
        mojo_per_unit = get_mojo_per_unit(wallet_type=wallet_type)
        final_amount: Optional[uint64] = uint64(int(Decimal(amount) * mojo_per_unit)) if amount else None
        res = await wallet_client.dao_create_proposal(
            wallet_id=wallet_id,
            proposal_type="spend",
            additions=additions,
            amount=final_amount,
            inner_address=address,
            asset_id=asset_id,
            vote_amount=vote_amount,
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        asset_id_name = asset_id if asset_id else "XCK"
        print(f"Created spend proposal for asset: {asset_id_name}")
        if push:
            print("Successfully created proposal.")
        print(f"Proposal ID: {res.proposal_id.hex()}")
        return res.transactions


async def create_update_proposal(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    fee: uint64,
    vote_amount: Optional[uint64],
    proposal_timelock: Optional[uint64],
    soft_close_length: Optional[uint64],
    attendance_required: Optional[uint64],
    pass_percentage: Optional[uint64],
    self_destruct_length: Optional[uint64],
    oracle_spend_delay: Optional[uint64],
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    new_dao_rules = {
        "proposal_timelock": proposal_timelock,
        "soft_close_length": soft_close_length,
        "attendance_required": attendance_required,
        "pass_percentage": pass_percentage,
        "self_destruct_length": self_destruct_length,
        "oracle_spend_delay": oracle_spend_delay,
    }
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_create_proposal(
            wallet_id=wallet_id,
            proposal_type="update",
            new_dao_rules=new_dao_rules,
            vote_amount=vote_amount,
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        if push:
            print("Successfully created proposal.")
        print(f"Proposal ID: {res.proposal_id.hex()}")
        return res.transactions


async def create_mint_proposal(
    root_path: pathlib.Path,
    wallet_rpc_port: Optional[int],
    fp: int,
    wallet_id: int,
    fee: uint64,
    amount: uint64,
    cat_target_address: str,
    vote_amount: Optional[int],
    cli_tx_config: CMDTXConfigLoader,
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    async with get_wallet_client(root_path, wallet_rpc_port, fp) as (wallet_client, fingerprint, config):
        res = await wallet_client.dao_create_proposal(
            wallet_id=wallet_id,
            proposal_type="mint",
            cat_target_address=cat_target_address,
            amount=amount,
            vote_amount=vote_amount,
            fee=fee,
            tx_config=cli_tx_config.to_tx_config(units["chik"], config, fingerprint),
            push=push,
            timelock_info=condition_valid_times,
        )

        if push:
            print("Successfully created proposal.")
        print(f"Proposal ID: {res.proposal_id.hex()}")
        return res.transactions



from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Optional

from chik_rs import AugSchemeMPL, G1Element, PrivateKey
from chikpos import DiskPlotter

from chik.daemon.keychain_proxy import KeychainProxy, connect_to_keychain_and_validate, wrap_local_keychain
from chik.plotting.util import Params, stream_plot_info_ph, stream_plot_info_pk
from chik.types.blockchain_format.proof_of_space import (
    calculate_plot_id_ph,
    calculate_plot_id_pk,
    generate_plot_public_key,
)
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import decode_puzzle_hash
from chik.util.keychain import Keychain
from chik.wallet.derive_keys import master_sk_to_farmer_sk, master_sk_to_local_sk, master_sk_to_pool_sk

log = logging.getLogger(__name__)


class PlotKeys:
    def __init__(
        self,
        farmer_public_key: G1Element,
        pool_public_key: Optional[G1Element],
        pool_contract_address: Optional[str],
    ):
        self.farmer_public_key = farmer_public_key
        self.pool_public_key = pool_public_key
        self.pool_contract_address = pool_contract_address

    @property
    def pool_contract_puzzle_hash(self) -> Optional[bytes32]:
        if self.pool_contract_address is not None:
            return decode_puzzle_hash(self.pool_contract_address)
        return None


class PlotKeysResolver:
    def __init__(
        self,
        farmer_public_key: Optional[str],
        alt_fingerprint: Optional[int],
        pool_public_key: Optional[str],
        pool_contract_address: Optional[str],
        root_path: Path,
        log: logging.Logger,
        connect_to_daemon: bool = False,
    ) -> None:
        self.farmer_public_key = farmer_public_key
        self.alt_fingerprint = alt_fingerprint
        self.pool_public_key = pool_public_key
        self.pool_contract_address = pool_contract_address
        self.root_path = root_path
        self.log = log
        self.connect_to_daemon = connect_to_daemon
        self.resolved_keys: Optional[PlotKeys] = None

    async def resolve(self) -> PlotKeys:
        if self.resolved_keys is not None:
            return self.resolved_keys

        keychain_proxy: Optional[KeychainProxy] = None
        try:
            if self.connect_to_daemon:
                keychain_proxy = await connect_to_keychain_and_validate(self.root_path, self.log)
            else:
                keychain_proxy = wrap_local_keychain(Keychain(), log=self.log)

            farmer_public_key: G1Element
            if self.farmer_public_key is not None:
                farmer_public_key = G1Element.from_bytes(bytes.fromhex(self.farmer_public_key))
            else:
                farmer_public_key = await self.get_farmer_public_key(keychain_proxy)

            pool_public_key: Optional[G1Element] = None
            if self.pool_public_key is not None:
                if self.pool_contract_address is not None:
                    raise RuntimeError("Choose one of pool_contract_address and pool_public_key")
                pool_public_key = G1Element.from_bytes(bytes.fromhex(self.pool_public_key))
            else:
                if self.pool_contract_address is None:
                    # If nothing is set, farms to the provided key (or the first key)
                    pool_public_key = await self.get_pool_public_key(keychain_proxy)

            self.resolved_keys = PlotKeys(farmer_public_key, pool_public_key, self.pool_contract_address)
        finally:
            if keychain_proxy is not None:
                await keychain_proxy.close()
        return self.resolved_keys

    async def get_sk(self, keychain_proxy: Optional[KeychainProxy] = None) -> Optional[PrivateKey]:
        sk: Optional[PrivateKey] = None
        if keychain_proxy:
            try:
                if self.alt_fingerprint is not None:
                    sk = await keychain_proxy.get_key_for_fingerprint(self.alt_fingerprint)
                else:
                    sk = await keychain_proxy.get_first_private_key()
            except Exception as e:
                log.error(f"Keychain proxy failed with error: {e}")
        else:
            sk_ent: Optional[tuple[PrivateKey, bytes]] = None
            keychain: Keychain = Keychain()
            if self.alt_fingerprint is not None:
                sk_ent = keychain.get_private_key_by_fingerprint(self.alt_fingerprint)
            else:
                sk_ent = keychain.get_first_private_key()

            if sk_ent:
                sk = sk_ent[0]
        return sk

    async def get_farmer_public_key(self, keychain_proxy: Optional[KeychainProxy] = None) -> G1Element:
        sk: Optional[PrivateKey] = await self.get_sk(keychain_proxy)
        if sk is None:
            raise RuntimeError(
                "No keys, please run 'chik keys add', 'chik keys generate' or provide a public key with -f"
            )
        return master_sk_to_farmer_sk(sk).get_g1()

    async def get_pool_public_key(self, keychain_proxy: Optional[KeychainProxy] = None) -> G1Element:
        sk: Optional[PrivateKey] = await self.get_sk(keychain_proxy)
        if sk is None:
            raise RuntimeError(
                "No keys, please run 'chik keys add', 'chik keys generate' or provide a public key with -p"
            )
        return master_sk_to_pool_sk(sk).get_g1()


async def resolve_plot_keys(
    farmer_public_key: Optional[str],
    alt_fingerprint: Optional[int],
    pool_public_key: Optional[str],
    pool_contract_address: Optional[str],
    root_path: Path,
    log: logging.Logger,
    connect_to_daemon: bool = False,
) -> PlotKeys:
    return await PlotKeysResolver(
        farmer_public_key, alt_fingerprint, pool_public_key, pool_contract_address, root_path, log, connect_to_daemon
    ).resolve()


async def create_plots(
    args: Params,
    keys: PlotKeys,
    use_datetime: bool = True,
    test_private_keys: Optional[list[PrivateKey]] = None,
) -> tuple[dict[bytes32, Path], dict[bytes32, Path]]:
    if args.tmp2_dir is None:
        args.tmp2_dir = args.tmp_dir
    assert (keys.pool_public_key is None) != (keys.pool_contract_puzzle_hash is None)
    num = args.num

    if keys.pool_public_key is not None:
        log.info(
            f"Creating {num} plots of size {args.size}, pool public key:  "
            f"{bytes(keys.pool_public_key).hex()} farmer public key: {bytes(keys.farmer_public_key).hex()}"
        )
    else:
        assert keys.pool_contract_puzzle_hash is not None
        log.info(
            f"Creating {num} plots of size {args.size}, pool contract address:  "
            f"{keys.pool_contract_address} farmer public key: {bytes(keys.farmer_public_key).hex()}"
        )

    tmp_dir_created = False
    if not args.tmp_dir.exists():
        args.tmp_dir.mkdir(parents=True, exist_ok=True)
        tmp_dir_created = True

    tmp2_dir_created = False
    if not args.tmp2_dir.exists():
        args.tmp2_dir.mkdir(parents=True, exist_ok=True)
        tmp2_dir_created = True

    args.final_dir.mkdir(parents=True, exist_ok=True)

    created_plots: dict[bytes32, Path] = {}
    existing_plots: dict[bytes32, Path] = {}
    for i in range(num):
        # Generate a random master secret key
        if test_private_keys is not None:
            assert len(test_private_keys) == num
            sk: PrivateKey = test_private_keys[i]
        else:
            sk = AugSchemeMPL.key_gen(bytes32.secret())

        # The plot public key is the combination of the harvester and farmer keys
        # New plots will also include a taproot of the keys, for extensibility
        include_taproot: bool = keys.pool_contract_puzzle_hash is not None
        plot_public_key = generate_plot_public_key(
            master_sk_to_local_sk(sk).get_g1(), keys.farmer_public_key, include_taproot
        )

        # The plot id is based on the harvester, farmer, and pool keys
        if keys.pool_public_key is not None:
            plot_id: bytes32 = calculate_plot_id_pk(keys.pool_public_key, plot_public_key)
            plot_memo: bytes = stream_plot_info_pk(keys.pool_public_key, keys.farmer_public_key, sk)
        else:
            assert keys.pool_contract_puzzle_hash is not None
            plot_id = calculate_plot_id_ph(keys.pool_contract_puzzle_hash, plot_public_key)
            plot_memo = stream_plot_info_ph(keys.pool_contract_puzzle_hash, keys.farmer_public_key, sk)

        if args.plotid is not None:
            log.info(f"Debug plot ID: {args.plotid}")
            # Check if args.memo is of type bytes and convert it to a string if so
            if isinstance(args.plotid, bytes):
                plot_str = args.plotid.hex()  # Convert bytes to hex string
            else:
                plot_str = args.plotid
            plot_id = bytes32.fromhex(plot_str)

        if args.memo is not None:
            log.info(f"Debug memo: {args.memo}")
            # Check if args.memo is of type bytes and convert it to a string if so
            if isinstance(args.memo, bytes):
                memo_str = args.memo.hex()  # Convert bytes to hex string
            else:
                memo_str = args.memo
            plot_memo = bytes.fromhex(memo_str)

        dt_string = datetime.now().strftime("%Y-%m-%d-%H-%M")

        if use_datetime:
            filename: str = f"plot-k{args.size}-{dt_string}-{plot_id}.plot"
        else:
            filename = f"plot-k{args.size}-{plot_id}.plot"
        full_path: Path = args.final_dir / filename

        if not full_path.exists():
            log.info(f"Starting plot {i + 1}/{num}")
            # Creates the plot. This will take a long time for larger plots.
            plotter: DiskPlotter = DiskPlotter()
            plotter.create_plot_disk(
                str(args.tmp_dir),
                str(args.tmp2_dir),
                str(args.final_dir),
                filename,
                args.size,
                plot_memo,
                plot_id,
                args.buffer,
                args.buckets,
                args.stripe_size,
                args.num_threads,
                args.nobitfield,
            )
            created_plots[plot_id] = full_path
        else:
            log.info(f"Plot {filename} already exists")
            existing_plots[plot_id] = full_path

    log.info("Summary:")

    if tmp_dir_created:
        try:
            args.tmp_dir.rmdir()
        except Exception:
            log.info(f"warning: did not remove primary temporary folder {args.tmp_dir}, it may not be empty.")

    if tmp2_dir_created:
        try:
            args.tmp2_dir.rmdir()
        except Exception:
            log.info(f"warning: did not remove secondary temporary folder {args.tmp2_dir}, it may not be empty.")

    log.info(f"Created a total of {len(created_plots)} new plots")
    for created_path in created_plots.values():
        log.info(created_path.name)

    return created_plots, existing_plots
